//   Copyright 2007, 2008 Nicolas Dubuit
//
//     This file is part of Immix.
//
//     Immix is free software: you can redistribute it and/or modify
//     it under the terms of the GNU General Public License as published by
//     the Free Software Foundation, either version 3 of the License, or
//     (at your option) any later version.
//
//     This program is distributed in the hope that it will be useful,
//     but WITHOUT ANY WARRANTY; without even the implied warranty of
//     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//     GNU General Public License for more details.
//
//     You should have received a copy of the GNU General Public License
//     along with this program.  If not, see <http://www.gnu.org/licenses/>.

#include "correlator.h"
#include "align.h"

#include <stdio.h>
#include <iostream>
#include <cmath>
using namespace std ;

#include <QCoreApplication>

#include <QFileDialog>
#include <QLabel>
#include <QMatrix>
#include <QPainter>

#include "QNImageDialog.h"
#include "immix.h"


int AlignmentManager::numThreads = 1 ;

AlignmentManager::AlignmentManager (ImmixWin*im) : QObject(0)
{
    // No finished threads yet.
    numFinishedTreads = 0 ;

    // Memorize immix window ptr.
    imw = im ;

	// Prepare the progress dialog.
	dialog = new AlignDialog (imw->displayList.size(), this, imw);

    // When the last thread finished, do the final processing.
    connect (this, SIGNAL(allThreadsFinished()),
             dialog, SLOT(finish()) ) ;
    connect (this, SIGNAL(allThreadsFinished()),
             this, SLOT(finalProcessing()) ) ;
}

AlignmentManager::~AlignmentManager (void)
{    
// Delete progress dialog.
    delete dialog ;
}


void AlignmentManager::startAlign (void)
{
	// Show the progress dialog.
    dialog->display ( imw->displayList.size() ) ;

    // The first unaligned image has number 1.
    nextUnalignedImage = 1 ;

    // No finished threads yet.
    numFinishedTreads = 0 ;

    // Calculate the correlation coefficient of the reference data with a window function.
    //   This gives for each value of the correlation function, the corresponding signal sum
    //   (or "window-average"). It is used for nomalization.
    // This work has only to be done once.
	QRect refRect1 = QRect (imw->mkr1.x()-ImmixWin::MkrSz/2, imw->mkr1.y()-ImmixWin::MkrSz/2,
	                                    ImmixWin::MkrSz, ImmixWin::MkrSz) ;
	QRect refRect2 = QRect (imw->mkr2.x()-ImmixWin::MkrSz/2, imw->mkr2.y()-ImmixWin::MkrSz/2,
	                                    ImmixWin::MkrSz,ImmixWin::MkrSz) ;
	QImage refImg1 = getImagePtr(0)->copy(refRect1) ;
	QImage refImg2 = getImagePtr(0)->copy(refRect2) ;

    RefWinCorrelator1 = new Correlator (ImmixWin::MkrSz, ImmixWin::FineFactor) ;
    RefWinCorrelator1->setRefDataSquare (refImg1) ;
    RefWinCorrelator1->setWindowTargetData () ;
    RefWinCorrelator1->calculateCorrelation () ;

    RefWinCorrelator2 = new Correlator  (ImmixWin::MkrSz, ImmixWin::FineFactor) ;
    RefWinCorrelator2->setRefDataSquare (refImg2) ;
    RefWinCorrelator2->setWindowTargetData () ;
    RefWinCorrelator2->calculateCorrelation () ;

    // These "window" correlations (the averages of signals over the shifted zones) are now
    // calculated. We only need the result of it now.
    RefWinCorrelator1->freeRefMemory () ;
    RefWinCorrelator1->freeFourierMemory () ;
    RefWinCorrelator2->freeRefMemory () ;
    RefWinCorrelator2->freeFourierMemory () ;

   // Create the worker threads, and start them.
    //   They will ask for work in due time.
    numThreads = QThread::idealThreadCount () ;
    cout << "Using " << numThreads << " threads." << endl ;
    for (int i=0 ; i<numThreads ; i++)
    {
        // Create a new worker thread.
        AligningThread* thread = new AligningThread (this) ;

        // Count thread when finished.
        connect (thread, SIGNAL(finished()), 
                 this, SLOT (threadFinished()),
		         Qt::QueuedConnection) ;

        // This will start aligning images.
        thread->start () ;
    }
    // We don't execute anything here : the work is currently happening in the threads.
}

int AlignmentManager::getNewImage (void)
{
    // The value can't be directly returned because the mutex has to be unlocked.
    int newImage ;

    // The following code is protected against multiple execution by different threads.
    nextUnalignedImageMutex.lock() ;
    
    // If there is a new image to align.
    if ( nextUnalignedImage<imw->displayList.size() )
    {
        newImage = nextUnalignedImage ;
        nextUnalignedImage++ ;
    }

    // If everything is already aligned. Any thread recieving such a result will stop.
    else
    {
        newImage = -1 ;
    }

    // End of protected code.
    nextUnalignedImageMutex.unlock() ;

    return newImage ;
}

QImage* AlignmentManager::getImagePtr (int num)
{
    return imw->displayList[num]->imgPtr ;
}

QString AlignmentManager::getImagePath (int num)
{
    return imw->displayList[num]->filePath[1] ;
}

void AlignmentManager::finalProcessing (void)
{
	// We still need to shift the first imagen, just like all others..
	QImage finale (getImagePtr(0)->width() + 2*ImmixWin::MkrSz, 
	               getImagePtr(0)->height() + 2*ImmixWin::MkrSz, 
	               QImage::Format_RGB32) ;
    QPainter finalePainter (&finale) ;
	finalePainter.fillRect (finalePainter.window(), QColor("black")) ;
	finalePainter.drawImage (QPoint(ImmixWin::MkrSz, ImmixWin::MkrSz), *getImagePtr(0)) ;

    // Write the new first image to disk.
	QString newPremsFileName = imw->displayList[0]->filePath[0]+ImmixWin::Suffix  ;
	finale.save (newPremsFileName, "jpg", ImmixWin::jpgQuality) ;

	// Copy EXIF metadata.
    ImagePtr imageExifRead = Exiv2::ImageFactory::open(imw->displayList[0]->filePath[0].toStdString());
    imageExifRead->readMetadata();
    ImagePtr imageExifWrite = Exiv2::ImageFactory::open(newPremsFileName.toStdString());
    imageExifWrite->setExifData(imageExifRead->exifData());
    imageExifWrite->writeMetadata();

    // Now display the new first image.
	imw->displayList[0]->ChangeFile (newPremsFileName) ;

	// Move the markers according to the new position of images (translated by MkrSz,MkrSz)
	imw->mkr1 += QPoint(ImmixWin::MkrSz,ImmixWin::MkrSz) ;
	imw->mkr2 += QPoint(ImmixWin::MkrSz,ImmixWin::MkrSz) ;
	imw->update() ;	

    // Give input focus to slider.
    imw->slider->setFocus (Qt::OtherFocusReason) ;

    // Allow "revert" button.
	imw->btnRevert->setEnabled (true) ;
}

void AlignmentManager::threadFinished (void)
{
    numFinishedTreadsMutex.lock() ;
    numFinishedTreads ++ ;
    if ( numFinishedTreads == numThreads)
        emit allThreadsFinished () ;
    numFinishedTreadsMutex.unlock() ; 
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


AligningThread::AligningThread (AlignmentManager*man)
{
// Don't create the aligner there, since we are still executing 
//   in the main thread.
    manager = man ;
}

void AligningThread::run(void) 
{
// Now that we are in the thread, we can create the aligner.
//   We create it in the stack so that it is destroyed at the end 
//   of the thread execution.
    Aligner aligner (manager) ;

// Connect signals and slots.
    connect (&aligner, SIGNAL(alignmentDone (int, double, double, double, double)),
             manager->dialog, SLOT(displayAlignmentData (int, double, double, double, double)),
             Qt::QueuedConnection) ;

// Loop asking for new images.
    int imgIdx ;
    while ( ( imgIdx=manager->getNewImage() ) != -1 )
    {
    // Align the given image.
        aligner.align(imgIdx) ;
    }
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


Aligner::Aligner (AlignmentManager* man)
{
    // Memorize parameters.
    manager = man ;
    MkrSz = ImmixWin::MkrSz ;
    FineFactor = ImmixWin::FineFactor ;

    // Prepare main correlators.
    Correlator1 = new Correlator (MkrSz, FineFactor) ;
    Correlator2 = new Correlator (MkrSz, FineFactor) ;
    
	// Get reference gray canal in the corresponding arrays for fourier transform.
	refRect1 = QRect (manager->imw->mkr1.x()-MkrSz/2,manager->imw->mkr1.y()-MkrSz/2,MkrSz,MkrSz) ;
	refRect2 = QRect (manager->imw->mkr2.x()-MkrSz/2,manager->imw->mkr2.y()-MkrSz/2,MkrSz,MkrSz) ;
	QImage refImg1 = manager->getImagePtr(0)->copy(refRect1) ;
	QImage refImg2 = manager->getImagePtr(0)->copy(refRect2) ;
    Correlator1->setRefData (refImg1) ; // This calculates the FFT.
    Correlator2->setRefData (refImg2) ; // This calculates the FFT.
    Correlator1->freeRefMemory () ;
    Correlator2->freeRefMemory () ;

    // Prepare correlations used for normalization.
    TargWinCorrelator1 = new Correlator (MkrSz, FineFactor) ;
    TargWinCorrelator2 = new Correlator (MkrSz, FineFactor) ;

	// The final image on which the transformed image is drawn at the right position.
	finale = QImage ((manager->getImagePtr(0))->width() + 2*MkrSz, 
	                 (manager->getImagePtr(0))->height() + 2*MkrSz, 
	                 QImage::Format_RGB32) ;
}

Aligner::~Aligner (void)
{
    delete Correlator1 ;
    delete Correlator2 ;
    delete TargWinCorrelator1 ;
    delete TargWinCorrelator2 ;
}

void Aligner::align (int idx)
{
	// Get a pointer on the image display.
	ImmixDisplay* imgDisp = manager->imw->displayList[idx] ;

	// Change the image filename.
	QString newFileName = imgDisp->filePath[0]+ImmixWin::Suffix  ;
	imgDisp->ChangeFile (newFileName) ;

	// Load the image from the old file name.
	currentImage.load(imgDisp->filePath[1]) ;
	QImage targImg1 = currentImage.copy(refRect1) ;
	QImage targImg2 = currentImage.copy(refRect2) ;
	Correlator1->setTargetData (targImg1) ;
	Correlator2->setTargetData (targImg2) ;

	// Calculate correlation.
	Correlator1->calculateCorrelation () ;
	Correlator2->calculateCorrelation () ;

	// Calculate the correlation coefficient of the reference data with a window function.
	//   This gives for each value of the correlation function, the corresponding signal sum
	//   (or "window-average").
	TargWinCorrelator1->setWindowRefData () ;
	TargWinCorrelator1->setTargetDataSquare (targImg1) ;
	TargWinCorrelator1->calculateCorrelation () ;

	TargWinCorrelator2->setWindowRefData () ;
	TargWinCorrelator2->setTargetDataSquare (targImg2) ;
	TargWinCorrelator2->calculateCorrelation () ;

	// Divide the correlations by the window-averages.
	//Correlator1.MultiplyByN() ;
	Correlator1->DivideByRootOf(*TargWinCorrelator1) ;
	Correlator1->DivideByRootOf(*(manager->RefWinCorrelator1)) ;

	//Correlator2.MultiplyByN() ;
	Correlator2->DivideByRootOf(*TargWinCorrelator2) ;
	Correlator2->DivideByRootOf(*(manager->RefWinCorrelator2)) ;

	// Find maximum of correlation function, the position of which being at pos(targ)-pos(ref).
	double dxCorMax1 = 0., dyCorMax1 = 0. ;
	double correlMin1, correlMax1 ;
	double IcorrelMin1, IcorrelMax1 ;
	Correlator1->findCorrelMax (dxCorMax1, dyCorMax1, correlMin1, correlMax1, IcorrelMin1, IcorrelMax1 );

	double dxCorMax2 = 0., dyCorMax2 = 0. ;
	double correlMin2, correlMax2 ;
	double IcorrelMin2, IcorrelMax2 ;
	Correlator2->findCorrelMax (dxCorMax2, dyCorMax2, correlMin2, correlMax2, IcorrelMin2, IcorrelMax2 );

	// image offset		
	cout << "Real : " << correlMin1 << " -> " << correlMax1 << " (" << dxCorMax1 << ", " << dyCorMax1 << ") , " 
	                  << correlMin2 << " -> " << correlMax2 << " (" << dxCorMax2 << ", " << dyCorMax2 << ")"
	     << endl ;
	
	// Comme on n'a pas de cisaillement et une dilatation homogène en x et en y, 
	//     m12=-m21 et m11 = m22.
	// Calcul des points de référence et de leur nouvelle position. 
	double x1,x2,y1,y2, xp1,xp2,yp1,yp2 ;
	x1 = manager->imw->mkr1.x() ; y1 = manager->imw->mkr1.y() ; 
	x2 = manager->imw->mkr2.x() ; y2 = manager->imw->mkr2.y() ;
	xp1 = x1 + dxCorMax1 ; yp1 = y1 + dyCorMax1 ;
	xp2 = x2 + dxCorMax2 ; yp2 = y2 + dyCorMax2 ;

	// Calcul des elements de la matrice (voir ci dessus pour le détail du calcul)
	//    D = (x1-x2)(x1-x2) + (y1-y2)(y1-y2)
	//    s = ( (xp1-xp2)(x1-x2)+(yp1-yp2)(y1-y2) )  /  D
	//    r = ( (yp1-yp2)(x1-x2) - (xp1-xp2)(y1-y2) )  /  D
	//    
	double m11, m12, m21, m22, deltax, deltay ;
	double denom = (x1-x2)*(x1-x2) + (y1-y2)*(y1-y2)  ;
	m11 = m22 =  ( (xp1-xp2)*(x1-x2) + (yp1-yp2)*(y1-y2) )  /  denom ;
	m12 =        ( (yp1-yp2)*(x1-x2) - (xp1-xp2)*(y1-y2) )  /  denom ;
	m21 = - m12 ;
	deltax = xp1 - m11 * x1 - m21 * y1 ;
	deltay = yp1 - m12 * x1 - m22 * y1 ;

	// Calculate the transformation matrix, and invert it.
	QMatrix transfo (m11,m12,m21,m22,deltax,deltay) ; 
	transfo = transfo.inverted() ;
	QMatrix realMatr = QImage::trueMatrix(transfo, currentImage.width(), currentImage.height()) ;

	// Transform the image using the inverted QMatrix.
	QImage transImg = currentImage.transformed (transfo, Qt::SmoothTransformation) ;

	// Move image to align it with the reference image..
	QPainter finalePainter (&finale) ;
	finalePainter.fillRect (finalePainter.window(), QColor("black")) ;
	finalePainter.drawImage (QPointF(MkrSz + (transfo.dx()-realMatr.dx()),
										MkrSz + (transfo.dy()-realMatr.dy())), 
								transImg) ;

	// Save the image using the new filename.
	finale.save (newFileName, "jpg", ImmixWin::jpgQuality) ;

	// Copy EXIF metadata.
	imageExifRead = Exiv2::ImageFactory::open(imgDisp->filePath[1].toStdString());
	imageExifRead->readMetadata();
	imageExifWrite = Exiv2::ImageFactory::open(newFileName.toStdString());
	imageExifWrite->setExifData(imageExifRead->exifData());
	imageExifWrite->writeMetadata();

    // Calculation of rotation and scale.
    double r = asin (transfo.m12()) ;
    double s = transfo.m11() / cos(r) ;

    // End of alignement.
    emit alignmentDone (idx, dxCorMax1, dyCorMax1, r*180./3.141592, s) ;
}











/*

Calcul des transformations
--------------------------

xa' = m11*xa + m21*ya + dx
ya' = m22*ya + m12*xa + dy
xb' = m11*xb + m21*yb + dx
yb' = m22*yb + m12*xb + dy
m12=-m21=r
m11=m22=s

xa' = s*xa - r *ya + dx
ya' = s*ya + r*xa + dy
xb' = s*xb - r*yb + dx
yb' = s*yb + r*xb + dy

dx = xa' - s*xa + r *ya
dy = ya' - s*ya - r*xa
0 = s*(xb-xa) - r*(yb-ya) + (xa' - xb')
0 = s*(yb-ya) + r*(xb-xa) +  (ya' - yb')

0 = - r*[(yb-ya)(yb-ya) + (xb-xa) (xb-xa)]  +  (xa' - xb')(yb-ya) - (ya' - yb') (xb-xa)
0 = s*[(xb-xa)(xb-xa) + (yb-ya) (yb-ya)]   +  (xa' - xb') (xb-xa) + (yb-ya) (ya' - yb')

r  =  [(xa' - xb')(yb-ya) - (ya' - yb') (xb-xa)]   /   [(yb-ya)(yb-ya) + (xb-xa) (xb-xa)]
s = - [(xa' - xb') (xb-xa) + (yb-ya) (ya' - yb')] /   [(xb-xa)(xb-xa) + (yb-ya) (yb-ya)]
dx = xa' 
       +  [(xa' - xb') (xb-xa) + (yb-ya) (ya' - yb')] /   [(xb-xa)(xb-xa) + (yb-ya) (yb-ya)] *xa 
       +   [(xa' - xb')(yb-ya) - (ya' - yb') (xb-xa)]   /   [(yb-ya)(yb-ya) + (xb-xa) (xb-xa)] *ya
dy =  ya'
       + [(xa' - xb') (xb-xa) + (yb-ya) (ya' - yb')] /   [(xb-xa)(xb-xa) + (yb-ya) (yb-ya)]*ya 
       -   [(xa' - xb')(yb-ya) - (ya' - yb') (xb-xa)]   /   [(yb-ya)(yb-ya) + (xb-xa) (xb-xa)]*xa



xp1 = s*x1 - r *y1 + dx
yp1 = s*y1 + r*x1 + dy
xp2 = s*x2 - r*y2 + dx
yp2 = s*y2 + r*x2 + dy

dx = xp1 - s*x1 + r *y1  
dx = xp2 - s*x2 + r *y2
dy = yp1 - s*y1 - r*x1 
dy = yp2 - s*y2 - r*x2 

(xp1-xp2) - s*(x1-x2) + r *(y1-y2) = 0
(yp1-yp2) - s*(y1-y2) - r *(x1-x2) = 0

(y1-y2)(x1-x2)

(xp1-xp2)(x1-x2) - s*(x1-x2)(x1-x2) + r *(y1-y2)(x1-x2) = 0
(yp1-yp2)(y1-y2) - s*(y1-y2)(y1-y2) - r *(x1-x2)(y1-y2) = 0

-> (xp1-xp2)(x1-x2)+(yp1-yp2)(y1-y2) -s*( (x1-x2)(x1-x2) + (y1-y2)(y1-y2) ) = 0
->    s = ( (xp1-xp2)(x1-x2)+(yp1-yp2)(y1-y2) )  /  ( (x1-x2)(x1-x2) + (y1-y2)(y1-y2)  )

(xp1-xp2)(y1-y2) - s*(x1-x2)(y1-y2) + r *(y1-y2)(y1-y2) = 0
(yp1-yp2)(x1-x2) - s*(y1-y2)(x1-x2) - r *(x1-x2)(x1-x2) = 0

-> (xp1-xp2)(y1-y2)-(yp1-yp2)(x1-x2) + r*( (y1-y2)(y1-y2) + (x1-x2)(x1-x2) ) = 0
->    r = ( (yp1-yp2)(x1-x2) - (xp1-xp2)(y1-y2) )  /  ( (y1-y2)(y1-y2) + (x1-x2)(x1-x2) )

*/


/*QImage* QImageFromData (int, int, fftw_complex*, double, double) ;

QImage* newQImageFromData (int szX, int szY, fftw_complex* data, double min, double max)
{
    double val ;
    QImage* img = new QImage (szX, szY, QImage::Format_RGB32) ;
    for (int x=0 ; x<szX ; x++)
      for (int y=0 ; y<szY ; y++)
      {
        val = ( data[(y+szY*x)][0] - min ) / ( max - min ) ;
        if (val<0.) val=0. ;
        if (val>1.) val=1. ;
        img->setPixel ( x, y, qRgb(int(val*255.), int(val*255.), int(val*255.) ) ) ;
      }
    return img ;
}*/

